@@ -56,6 +56,7 @@ from __future__ import print_function
 
 import sys
 import os
+import re
 import signal
 import socket
 import select
@@ -73,7 +74,7 @@ from types import GeneratorType, ModuleType
 from collections import deque
 from binascii import a2b_hex, b2a_hex
 
-VERSION = '0.14'
+VERSION = '0.15'
 
 PYTHON3 = sys.version_info >= (3, 0)
 if PYTHON3:
@@ -108,7 +109,7 @@ __all__ = ['AutoThreadingClient', 'Client', 'ClientGroup', 'Connection',
            'define_common', 'definitions_module', 'eval_multi', 'exec_return',
            'execblock_multi', 'generate_key', 'key_from_options',
            'make_server_class', 'make_singleton', 'read_key_file',
-           'write_key_file',]
+           'write_key_file', 'parse_connection_spec']
 
 ## Special value for TID which causes thread to exit after one request
 SINGLE = -1
@@ -275,16 +276,43 @@ def _create_attr_wrapper(name):
 
     return property(get, set, delete)
 
+def _create_cached_attr_wrapper(name):
+    non_existant = object()
+    def get(self):
+        d = self.__dict__
+        val = d.get(name, non_existant)
+        if val is not non_existant:
+            return val
+        val = self._connection.send_request(0, 'getattr', (self, name))
+        d[name] = val
+        return val
+
+    def set(self, value):
+        self.__dict__.pop(name, None)
+        rv = self._connection.send_request(0, 'setattr', (self, name, value))
+        return rv
+
+    def delete(self):
+        self.__dict__.pop(name, None)
+        return self._connection.send_request(0, 'delattr', (self, name))
+
+    return property(get, set, delete)
+
 class RemoteProxyMeta(type):
     def __new__(mcs, clsname, bases, typedict):
         proxy_methods = typedict.setdefault('proxy_methods', ())
         proxy_attrs = typedict.setdefault('proxy_attrs', ())
+        proxy_attrs_cache = typedict.setdefault('proxy_attrs_cache', ())
+
         for name in proxy_methods:
             typedict[name] = _create_wrapper_method(name)
 
         for name in proxy_attrs:
             typedict[name] = _create_attr_wrapper(name)
         
+        for name in proxy_attrs_cache:
+            typedict[name] = _create_cached_attr_wrapper(name)
+        
         return super(RemoteProxyMeta, mcs).__new__(mcs, clsname, bases, typedict)
 
 RemoteProxyBase = RemoteProxyMeta('RemoteProxyBase', (object,), {'__module__': __name__})
@@ -667,23 +695,25 @@ class Client(Connection):
     _sendiv = b'CLNT'
     _recviv = b'SRVR'
     
-    def __init__(self, key, sock_or_host=None, port=DEFAULT_PORT,
+    def __init__(self, key, host=None, port=DEFAULT_PORT,
                  definitions='easycluster.remote_code'):
         '''Create a new Client object.
 
         `key` is the HMAC key to use.
 
-        `sock_or_host` is either an open socket or a host name, or None. If None
-        is given, you must call connect() or set_socket() before using the
+        `host` is a host name or IP address to connect to. If `host` is not
+        specified, you must call connect() or set_socket() before using the
         Client instance.
+        
+        `port` is the TCP port to use, if a host name is given..
 
-        `port` is the TCP port to use, if a host name is given.
+        `sock` is an open socket, or None.
 
         `definitions` is the name of the dynamic module where common definitions
         are stored. This must be the same name passed to define_common() or
         definitions_module().
         '''
-        Connection.__init__(self, key)
+        super(Client, self).__init__(key)
 
         self._next_request_id = 0
         self._pending_requests = None
@@ -713,10 +743,17 @@ class Client(Connection):
         self._init_data['definitions_source'] = []
         self._peer_version = None
 
-        if hasattr(sock_or_host, 'send'):
-            self.set_socket(sock_or_host)
-        elif sock_or_host is not None:
-            self.connect(sock_or_host, port)
+        if host is not None:
+            self.connect(host, port)
+
+    @classmethod
+    def from_spec(cls, spec, default_port=DEFAULT_PORT, default_key=None, warn=True):
+        '''Convenience method which creates a new instance from a connection spec.
+        See parse_connection_spec for the meaning of the parameters.
+        '''
+
+        params = parse_connection_spec(spec, default_port, default_key, (), warn)
+        return cls(**params)
 
     def _init_thread(self):
         pass
@@ -1087,7 +1124,53 @@ class Client(Connection):
             nbr = nbr()
             if nbr:
                 nbr.set_response(error)
-        
+
+class _ATCThread(threading.Thread):
+    def __init__(self, atc):
+        threading.Thread.__init__(self)
+        self.atc = atc
+        self.daemon = True
+
+    def run(self):
+        '''Read responses from the server.'''
+        atc = self.atc
+        next_wake_time = 0
+        fd = atc.fileno()
+        atc._got_keepalive()
+        closeargs = ()
+        try:
+            while True:
+                ctime = time.time()
+                timeout = atc._keep_alive_timeout
+                if timeout is not None:
+                    if ctime >= atc._last_keep_alive + timeout:
+                        closeargs = ('Remote server not responding',)
+                        return
+
+                if ctime >= next_wake_time:
+                    next_wake_time = ctime + 1
+                    # Wake up any threads that are waiting so they can check if they've
+                    # been interrupted.
+                    with atc._response_lock:
+                        atc._have_response.notify_all()
+                    continue
+                rd, _, _ = select.select([fd], (), (), next_wake_time - ctime)
+                if rd:
+                    if atc._read_data(False):
+                        ## Check thread status after every valid message received
+                        atc.check_threads()
+        except (socket.error, select.error) as e:
+            if e[0] != errno.ECONNRESET and e[0] != errno.EBADF:
+                raise
+        except EOFError:
+            ## Swallow this error so it isn't printed. Since we call close(),
+            ## any outstanding or future requests will fail.
+            pass
+        except Exception as e:
+            closeargs = (e,)
+            raise
+        finally:
+            atc.close(*closeargs)
 
 class AutoThreadingClient(Client, threading.Thread):
     '''A client that automatically creates unique threads on the server for each
@@ -1104,12 +1187,10 @@ class AutoThreadingClient(Client, threading.Thread):
     called. This class is also incompatible with ClientGroup.
     '''
     
-    def __init__(self, key, sock_or_host=None, port=DEFAULT_PORT,
+    def __init__(self, key, host=None, port=DEFAULT_PORT,
                  definitions='easycluster.remote_code', keep_alive_interval=10, 
                  keep_alive_timeout=60):
         '''Creates a new AutoThreadingClient. Arguments are the same as Client.'''
-        threading.Thread.__init__(self)
-        self.daemon = True
 
 
         ## use pseudo-thread IDs instead of sending real ones
@@ -1120,13 +1201,16 @@ class AutoThreadingClient(Client, threading.Thread):
         self._keep_alive_interval = keep_alive_interval
         self._keep_alive_timeout = keep_alive_timeout
         self._last_keep_alive = None
+        self._wrthread = None
 
-        Client.__init__(self, key, sock_or_host, port, definitions)
+        Client.__init__(self, key, host, port, definitions)
 
         self._init_data['keep_alive_interval'] = keep_alive_interval
 
     def _init_thread(self):
-        self.start()
+        thread = _ATCThread(self)
+        self._wrthread = weakref.ref(thread)
+        thread.start()
 
     def _got_keepalive(self):
         self._last_keep_alive = time.time()
@@ -1145,44 +1229,6 @@ class AutoThreadingClient(Client, threading.Thread):
         for id in done:
             self.stop_remote_thread(id)
 
-    def run(self):
-        '''Read responses from the server.'''
-        next_wake_time = 0
-        fd = self.fileno()
-        self._got_keepalive()
-        closeargs = ()
-        try:
-            while True:
-                ctime = time.time()
-                timeout = self._keep_alive_timeout
-                if timeout is not None:
-                    if ctime >= self._last_keep_alive + timeout:
-                        closeargs = ('Remote server not responding',)
-                        return
-
-                if ctime >= next_wake_time:
-                    next_wake_time = ctime + 1
-                    # Wake up any threads that are waiting so they can check if they've
-                    # been interrupted.
-                    with self._response_lock:
-                        self._have_response.notify_all()
-                    continue
-                rd, _, _ = select.select([fd], (), (), next_wake_time - ctime)
-                if rd:
-                    if self._read_data(False):
-                        ## Check thread status after every valid message received
-                        self.check_threads()
-        except (socket.error, select.error) as e:
-            if e[0] != errno.ECONNRESET and e[0] != errno.EBADF:
-                raise
-        except EOFError:
-            ## Swallow this error so it isn't printed. Since we call close(),
-            ## any outstanding or future requests will fail.
-            pass
-        except Exception as e:
-            raise
-        finally:
-            self.close(*closeargs)
 
     def read_response(self, *args):
         '''Do not call this method on AutoThreadingClient.'''
@@ -1218,12 +1264,26 @@ class AutoThreadingClient(Client, threading.Thread):
 
     def close(self, *a):
         '''Close the connection and stop the processing thread.'''
+        # It's possible that this function could be called recursively, so test
+        # if sock is None before we grab the lock.
+        if self._sock is None:
+            return
+
+        try:
+            thread = self._wrthread()
+        except TypeError:
+            # 'NoneType' is not callable.
+            thread = None
+
         ## This will shut down the socket and cause the reading thread to exit
         ## cleanly.
         with self._response_lock:
             Client.close(self, *a)
             self._have_response.notify_all()
 
+        if thread and thread != threading.current_thread():
+            thread.join()
+
 class ClientGroup:
     '''Represents a set of Client objects that can be queried as a group.'''
 
@@ -1381,6 +1441,103 @@ def key_from_options(opts, warn=True):
         print('SECURITY WARNING: running with blank key.', file=sys.stderr)
     return key
 
+def parse_connection_spec(spec, default_port=DEFAULT_PORT, default_key=b'', custom_options=(), warn=True):
+    '''Parse connection parameters from a string. This is useful if
+    you need to provide options to connect to more than one server.
+    
+    The connection spec looks like this:
+
+    host[:port][:opt=val]...
+    
+    The 'host' can be a hostname, IPv4 address, or bracketed IPv6 address. If
+    'port' is specified, it overrides the 'default_port' parameter.
+
+    The key is determined by specifying either the 'kf' or 'key' options. If
+    'key' is specified and 'warn' is True, a warning will be printed to
+    STDERR. If the connection spec comes does not come from the command line or
+    an environment variable, it's safe to set 'warn' to False. If no key is
+    specified, 'default_key' is used.
+
+    The 'custom_options' parameter allows the user to specify extra
+    program-specific per-client options. It can be a sequence of allowed
+    options, or True to allow any arbitrary options.
+
+    The return value is a dictionary of parameters which can be passed to the
+    Client or AutoThreadingClient constructors:
+
+     >>> params = easycluster.parse_connection_spec(sys.argv[1])
+     >>> client = easycluster.Client(**params)
+
+    If 'custom_options' is specified, two dictionaries are returned instead of
+    one:
+
+     >>> spec = 'example.com:1234:kf=secret.key:foo=bar'
+     >>> params, custom = easycluster.parse_connection_spec(spec, custom_options=['foo'])
+     >>> params
+     {'host': 'example.com', 'port': 1234, 'key': '...'}
+     >>> custom
+     {'foo': 'bar'}
+
+    Example connection specifications:
+    
+    'example.com'                    # Bare host name; default port, default key
+    '192.0.2.1:9999'                 # IPv4 address on non-standard port
+    '[2001:db8::2]'                  # IPv6 addresses must be in brackets
+    'example.com:kf=secret.key'      # Using a key from a file
+    'example.com:9999:key=s3cret'    # Using a key directly, with non-standard port
+
+    Raises ValueError if any option is duplicated, or options are
+    malformed. Raises IOError if a keyfile is specified, but can't be opened.
+    '''
+    
+    # Extract the host. Process a bracketed IPv6 address if it exists.
+    m = re.match(r'\[([0-9a-fA-F:]+)\]:(.*)', spec)
+    if m:
+        host = m.group(1)
+        rest = m.group(2)
+    else:
+        host, _, rest = spec.partition(':')
+        
+    # If the first value is a port number, parse it.
+    lst = rest.split(':')
+    if re.match(r'^\d+$', lst[0]):
+        port = int(lst.pop(0))
+    else:
+        port = default_port
+
+    key = None
+    custom = {}
+    for option in lst:
+        if not option:
+            continue
+        name, sep, val = option.partition('=')
+        if not sep:
+            raise ValueError('Malformed option %r in connection spec %r' % (option, spec))
+        if name == 'kf' or name == 'keyfile' or name == 'key':
+            if key is not None:
+                raise ValueError('Duplicate key %r in connection spec %r' % (option, spec))
+            if name == 'key':
+                if warn:
+                    print('SECURITY WARNING: using key from command line.', file=sys.stderr)
+                key = decode_key(val)
+            else:
+                key = read_key_file(val)
+        elif custom_options is True or name in custom_options:
+            if name in custom:
+                raise ValueError('Duplicate option %r in connection spec %r' % (option, spec))
+            custom[name] = val
+        else:
+            raise ValueError('Unknown option %r in connection spec %r' % (name, spec))
+
+    if key is None:
+        key = default_key
+
+    params = {'host': host, 'port': port, 'key': key}
+
+    if custom_options:
+        return params, custom
+    else:
+        return params
 
 ###############################
 # Internal classes
